/***************************************************************/
/*                    sjisngram.c                              */
/*　sjis 전각문자로 기술된  파일에 대한                        */
/*              n-gram을 작성하는 프로그램                     */
/* 입력은 표준입력으로 주고 출력은 표준출력으로 출력합니다     */
/* 사용법　                                                    */
/*  $./sjisngram (N 값）< (입력파일명）> (출력파일명)          */
/*　입력파일에는 텍스트파일을 지정합니다                       */
/*　출력파일에는 ngram을 출력합니다                            */
/***************************************************************/

/* 헤더 파일 포함 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* 기호상수 정의 */
#define N 32 /* ngram 최대길이*/

/* 함수의 프로토타입 선언 */
void  setlastch(int n,char chr,char lastch[]) ;
/* 과거 문자열 저장 */
void  printngram(int n,char lastch[]) ;/* ngram 출력 */

/*******************/
/*    main()함수   */
/*******************/
int main(int argc,char *argv[])
{
 int i ; /* 반복의 제어문자 */
 int n ; /* ngram의 길이 */
 char lastch[N]={' '} ; /* 과거의 값을 보존하는 변수 */
 char chr ;/* 입력된 문자 */
 unsigned char d ;/* 비교용 */

 /* 커맨드라인 인수 점검 */
 if(argc!=2){
  fprintf(stderr,"사용법　$./sjisngram (N 값)"
                 " < (입력파일명）> (출력파일명)\n") ;
  exit(1) ;
 }

 /* N 값 설정 */
 if(((n=atoi(argv[1]))<1)||(n>=N/2)){
  fprintf(stderr,"N 값이 부적절합니다.\n") ;
  exit(1) ;
 }

 /* sjis에서는 2바이트로 한 문자를 표현한다 */
 n=n*2 ;

 /* ngram 작성의 반복 */
 while((chr=getchar())!=EOF){
  d=(unsigned char) chr ;
  if(((d>0x7f)&&(d<0xa0))||(d>0xdf)&&(d<0xf0)){
   /*sjis의 2바이트 문자 */
   /* 과거 문자열 보존 */
   setlastch(n,chr,lastch) ;
   chr=getchar() ;
   setlastch(n,chr,lastch) ;

   /* ngram 출력 */
   printngram(n,lastch) ;
  }
 }

 return 0 ;
}

/************************/
/*  setlastch()함수     */
/*  과거 문자열 보존    */
/************************/
void  setlastch(int n,char chr,char lastch[])
{
 int i ;

 /* 과거의 데이터를 차례로 보낸다 */
 for(i=n-2;i>=0;--i)
  lastch[i+1]=lastch[i] ;

 /* 새로운 데이터 값을 세팅 */
 lastch[0]=chr ;
}

/************************/
/*  printngram()함수    */
/*  ngram 출력          */
/************************/
void  printngram(int n,char lastch[])
{
 int i ;
 
 /* ngram 출력 */
 for(i=n-1;i>=0;--i)
  printf("%c",lastch[i]) ;
 printf("\n") ;
}

