/*****************************************************/
/*                     gaml.c                        */
/*　SGA에 기초한 지식획득 프로그램                   */
/* 사용방법　                                        */
/*   $./gaml < (입력파일명）                         */
/*입력파일에서는 속성과 카테고리를 0/1로 기술합니다  */
/*****************************************************/

/* 헤더 파일 포함 */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

/* 기호 상수 정의 */
#define POOLSIZE 30 /* 풀 사이즈 */
#define LASTG 200    /* 중단 세대 */
#define MRATE 0.01  /* 돌연변이의 확률 */
#define SEED 65535    /* 난수의 시드 */
#define YES 1   /* yes에 대한 정수*/
#define NO 0    /* no에 대한 정수 */
#define TEXTLENGTH 256 /* 텍스트의 최대 길이 */
#define DATANO 256   /* 학습 데이터 세트의 최대 개수 */
#define ATTNO 4 /* 속성값의 개수 */
#define TERMNO 3 /* 논리식의 논리합 항의 개수 */

/* 함수의 프로토타입 선언 */
int evalfit(int gene[]) ; /* 적응도 계산 */
void mating(int pool[POOLSIZE][TERMNO*ATTNO*2]) ; /* 교차 */
void mutation(int pool[POOLSIZE][TERMNO*ATTNO*2]) ; /* 돌연변이 */
void printp(int pool[POOLSIZE][TERMNO*ATTNO*2]) ; /* 결과출력 */
void initpool(int pool[POOLSIZE][TERMNO*ATTNO*2]) ;
                         /* 초기집단 생성 */
int rndn() ; /*n미만 난수 생성   */
int notval(int v) ;/* 진리값 반전 */
int select(int roulette[POOLSIZE],int totalfitness) ;
                    /* 선택 */
void copypool(int pool[POOLSIZE][TERMNO*ATTNO*2]
          ,int nextpool[POOLSIZE][TERMNO*ATTNO*2]) ;
                 /* 다음 세대의 풀을 복사 */
void crossing(int m[],int p[],int c1[],int c2[]) ;
                /* 특정한 두 유전자의 교차 */
int readdata(int dataset[DATANO][ATTNO+1]) ;
                     /* 학습 데이터 읽어들이기 */
int evals(int formula[TERMNO*ATTNO*2]
      ,int subset[ATTNO+1]); /* 특정 사례에 대한 평가 */

/* 문제의 수치를 저장하는 배열 */
int no ; /* 학습 데이터 세트의 개수 */
int dataset[DATANO][ATTNO+1]={0} ;
       /* 학습 데이터 세트의 속성과 카테고리를 저장 */


/*******************/
/*    main()함수   */
/*******************/
int main(int argc,char *argv[])
{
 int pool[POOLSIZE][TERMNO*ATTNO*2] ; /* 유전자 풀 */
 int generation;/* 현재의 세대수 */

 /* 난수의 초기화 */
 srand(SEED) ;

 /* 학습 데이터 세트 읽어들이기 */
 no=readdata(dataset) ;

 /* 초기집단 생성 */
 initpool(pool) ;

 /* 중단세대까지 반복 */
 for(generation=0;generation<LASTG;++generation){
  printf("%d세대\n",generation) ;
  mating(pool) ;/* 교차 */
  mutation(pool) ;/* 돌연변이 */
  printp(pool) ;/* 결과출력 */
 }
 
 return 0 ;
}


/************************/
/*   select()함수       */
/*       선택           */
/************************/
int select(int roulette[POOLSIZE],int totalfitness)
{
 int i ;/* 반복 제어 변수 */
 int ball ;/* 공（선택 위치의 수치）*/
 int acc=0 ;/*적응도의 적산값 */

 ball=rndn(totalfitness) ;
 for(i=0;i<POOLSIZE;++i){
  acc+=roulette[i] ;/* 적응도를 적산 */
  if(acc>ball) break ;/* 대응하는 유전자 */
 }
   
 return i ;

}


/************************/
/*   mating()함수       */
/*        교차          */
/************************/
void mating(int pool[POOLSIZE][TERMNO*ATTNO*2])
{
 int i ;/* 반복 제어 변수 */
 int totalfitness=0 ;/* 적응도의 합계 */
 int nextpool[POOLSIZE][TERMNO*ATTNO*2] ;/* 자식 세대 */
 int roulette[POOLSIZE] ;/* 적응도를 저장 */
 int mama,papa ;/* 부모 유전자의 번호 */

 /* 룰렛 작성 */
 for(i=0;i<POOLSIZE;++i){
  roulette[i]=evalfit(pool[i]) ;
  /* 적응도의 합계치 계산 */
  totalfitness+=roulette[i] ;
 }

 /* 선택과 교차를 반복한다 */
 for(i=0;i<POOLSIZE/2;++i){
  do{/* 부모의 선택 */
   mama=select(roulette,totalfitness) ;
   papa=select(roulette,totalfitness) ;
  }while(mama==papa) ;/* 중복 제거 */
  /* 특정한 두 유전자의 교차 */
  crossing(pool[mama],pool[papa]
            ,nextpool[i*2],nextpool[i*2+1])  ;
 }

 /* 다음 세대의 풀을 복사 */
 copypool(pool,nextpool) ;

}
/************************/
/*  crossing()함수      */
/* 특정 두 유전자의 교차*/
/************************/
void crossing(int m[],int p[],int c1[],int c2[])
{
 int j ;/* 반복 제어 변수 */
 int cp ;/* 교차하는 점 */

 /* 교차점의 결정 */
 cp =rndn(TERMNO*ATTNO*2) ;

 /* 유전자의 전반부분 복사 */
 for(j=0;j<cp;++j){
  c1[j]=m[j] ;
  c2[j]=p[j] ;
 }
 
 /* 유전자의 후반부분 복사 */
 for(;j<TERMNO*ATTNO*2;++j){
  c2[j]=m[j] ;
  c1[j]=p[j] ;
 }
}

/************************/
/*  copypool()함수      */
/*다음 세대의 풀을 복사 */
/************************/
void copypool(int pool[POOLSIZE][TERMNO*ATTNO*2]
          ,int nextpool[POOLSIZE][TERMNO*ATTNO*2])
{
 int i,j ;/* 반복 제어 변수 */

 for(i=0;i<POOLSIZE;++i)
  for(j=0;j<TERMNO*ATTNO*2;++j)
   pool[i][j]=nextpool[i][j] ;
}

/************************/
/*   evalfit()함수      */
/*   적응도 계산        */
/************************/
int evalfit(int g[])
{
 int i ;/* 풀 제어 변수 */
 int point=0 ;/* 정해의 개수 */

 /* 학습 세트의 각 사례에 논리식을 적용 */
 for(i=0;i<no;++i){
  point+=evals(g,dataset[i]) ;
 }

 return point*100/no ;/* 평가값을 반환한다 */
}

/************************/
/*    evals()함수       */
/*특정 사례에 대한 평가 */
/************************/
int evals(int formula[TERMNO*ATTNO*2]
  ,int subset[ATTNO+1])
{
 int orval,andval,result ;/* 계산의 도중경과값 */
 int i,j ;/* 풀 제어 변수 */

 andval=YES ;
 /* 논리곱으로 항을 연결한 식 전체의 평가 */
 for(i=0;i<TERMNO;++i){
  orval=NO ;
  /* 논리합으로 연결한 항의 평가 */
  for(j=0;j<ATTNO*2;++j){
   if((j%2)==0) orval+=subset[j/2]*formula[i*ATTNO*2+j] ;
   else orval+=notval(subset[j/2])*formula[i*ATTNO*2+j] ;
  }
  if(orval==NO) andval=NO ;
  else andval*=YES ;
 }

 if(andval==subset[ATTNO]) return 1 ;/* 결과가 일치 */
 else return 0 ;/* 불일치 */
}

/***********************/
/*   printp()함수      */
/*   결과출력          */
/***********************/
void printp(int pool[POOLSIZE][TERMNO*ATTNO*2])
{
 int i,j ;/* 반복 제어 변수 */
 int fitness ;/* 적응도 */
 double totalfitness=0 ;/* 적응도의 합계 */
 int elite,bestfit=0 ;/* 엘리트 유전자 처리용 변수 */ 

 for(i=0;i<POOLSIZE;++i){
  for(j=0;j<TERMNO*ATTNO*2;++j)
   printf("%1d",pool[i][j]) ;
  fitness=evalfit(pool[i]) ;
  printf("\t%d\n",fitness) ;
  if(fitness>bestfit){/* 엘리트 해 */
   bestfit=fitness ;
   elite=i ;
  }
  totalfitness+=fitness ;
 }
 /* 엘리트 해의 적응도 출력 */
 printf("%d\t%d   ",elite,bestfit) ;
 /* 평균의 적응도 출력 */
 printf("%lf\n",totalfitness/POOLSIZE) ;
}


/***********************/
/*   initpool()함수    */
/*   초기집단 생성     */
/***********************/
void initpool(int pool[POOLSIZE][TERMNO*ATTNO*2])
{
 int i,j ;/* 반복 제어 변수 */

 for(i=0;i<POOLSIZE;++i)
  for(j=0;j<TERMNO*ATTNO*2;++j)
   pool[i][j]=rndn(2) ;

}


/************************/
/*     rndn()함수       */
/*    n미만 난수 생성 */
/************************/
int rndn(int l)
{
 int rndno ;/* 생성한 난수 */

 while((rndno=((double)rand()/RAND_MAX)*l)==l) ;

 return rndno;
}

/***********************/
/*   mutation()함수    */
/*   돌연변이          */
/***********************/
void mutation(int pool[POOLSIZE][TERMNO*ATTNO*2])
{
 int i,j ;/* 반복 제어 변수 */

 for(i=0;i<POOLSIZE;++i)
  for(j=0;j<TERMNO*ATTNO*2;++j)
   if((double)rndn(100)/100.0<=MRATE)
    /* 반전 돌연변이 */
    pool[i][j]=notval(pool[i][j]) ;

}


/************************/
/*   notval()함수       */
/*   진리값의 반전      */
/************************/
int notval(int v)
{
 if(v==YES) return NO ;
 else return YES ;
}

/************************/
/*   readdata()함수     */
/*학습 데이터 읽어들이기*/
/************************/
int readdata(int dataset[DATANO][ATTNO+1])
{
 int no=0 ; /* 학습 데이터 세트 개수 */
 int att=0 ; /*학습 데이터 세트 속성 */

 /*데이터 입력 */
 while(scanf("%d",&dataset[no][att])!=EOF){
  ++ att ;
  if(att>ATTNO){/* 다음 데이터 */
   att=0 ;
   ++no ;
  }
 }
 return no ;
}
